#include <plib.h>
#include "MyData.h"
#include "MyLibrary.h"
#include "NU32.h"

// function for initializing the ADC module
void ADCInitManual(void) {
  AD1PCFG = 0xF7FF; // Set AN11 to be an input
  AD1CON3bits.ADCS = 3; // ADC clock period is Tad = 2 * ADCS * Tpb
  AD1CON1bits.ADON = 1; // turn on A/D converter
}

// This function is used for manually reading pins on the ADC module
int ADCManualRead(int pin) {
  unsigned int elapsed, finishtime;
  // Connect the pin to the multiplexer
  AD1CHSbits.CH0SA = pin;
  // Start sampling
  AD1CON1bits.SAMP = 1;
  elapsed = ReadCoreTimer();
  finishtime = elapsed + SAMPLE_TIME;
  while (ReadCoreTimer() < finishtime);
  AD1CON1bits.SAMP = 0;
  while (!AD1CON1bits.DONE);
  return (ADC1BUF0);
}

// Initialize PWM on OC1
void StartPWM(int prescaler, int period) {
  // setup PWM on OC1
  OC1CONbits.OCM = 6; // PWM mode without fault pin
  OC1RS = 0; // initial PWM match
  OC1R = 0; // initial PWM match
  T2CON = prescaler << 4; // setup Timer2 prescaler
  PR2 = period; // Timer2 period match for PWM
  T2CONbits.ON = 1; // turn Timer2 on
  OC1CONbits.ON = 1; // turn PWM on
  // setup direction and idle pins
  TRISACLR = 0xC000; // set A14 and A15 to be outputs
}

// used for setting the direction of the DC motor
void CW(void) {
  LATASET = 0x8000;
  LATACLR = 0x4000;
}

// used for setting the direction of the DC motor
void CCW(void) {
  LATACLR = 0x8000;
  LATASET = 0x4000;
}


// used for stopping the DC motor
void STOP(void) {
  LATACLR = 0xC000;
  PWM_Set_DC(0);
}

// following is used for setting the duty cycle as a percent
void PWM_Set_DC(int dc) {
  // first let's convert the percent to a number of counts
  duty1 = (int) ((dc * (TMR2_PR + 1)) / 100);
  OC1RS = duty1;
}


// initializing SPI module for talking to encoder
void InitEncoderSPI(void) {
  int data = 0;
  SPI4CON = 0; // stop and reset SPI4
  data = SPI4BUF; // clear the rx buffer
  SPI4BRG = 0x4; // bit rate to 8MHz, SPI4BRG = 80000000/(2*desired)-1
  SPI4STATCLR = 0x40; // clear overflow
  SPI4CON = 0x100086A0; // MSSEN ON enables SS, SPI ON, 16 bit xfer, SMP=1, MSTR
}

// function for getting the current encoder counts
long long GetEncoder(int reset) {
  static int curr=0, last=0, data=0;
  static long long total_count=0;
  SPI4BUF = 1; // request the encoder position
  while (!SPI4STATbits.SPIRBF);
  data = SPI4BUF; // garbage was transfered over, ignore it
  SPI4BUF = 5; // write garbage, but the corresponding read will have the data
  while (!SPI4STATbits.SPIRBF);

  curr = SPI4BUF;
  // now we need to account for rollover:
  if (curr >= last) {
    // assume CCW
    if (curr-last > MAX_COUNTS_PER_SECOND)
      // negative rollover
      total_count -= 65536+last-curr;
    else
      // no rollover
      total_count += curr-last;
  } else {
    // assume CW
    if (last-curr > MAX_COUNTS_PER_SECOND)
      // positive rollover
      total_count += 65536+curr-last;
    else
      total_count -= last-curr;
  }
  last = curr;

  if (reset)
    total_count = 0;
  
  return (total_count);
}


// the following function is used to setup the high-frequency current
// control loop.  It initializes timer 4 to interrupt at
// CURR_CONTROL_FREQ and it initializes the ADC module for manual
// reading
void InitCurrentControl(int prescaler, int period) {
  // first initialize the ADC port
  ADCInitManual();
  // Now configure timer 4 to function properly
  TMRInitFixedFreq(prescaler, period);
}


// initialize a timer to interrupt at a fixed frequency
void TMRInitFixedFreq(int prescaler, int period) {
  T4CON = prescaler << 4; // prescaler 1
  T4CONCLR = 0x2; // internal pulses
  T4CONCLR = (1 << 13 | 1 << 7 | 1 << 3); // continue in idle, disable gated
  PR4 = period; // set period
  IPC4 = 0b11100; // priority 7, 0 subpriority
  IFS0bits.T4IF = 0; // clear flag
  T4CONSET = 1 << 15; // turn on timer
  IEC0bits.T4IE = 1; // enable interrupts
}

// this function is for setting up the low-frequency motion control loop
void InitMotionControl(int prescaler, int period) {
  T5CON = prescaler << 4; // prescaler
  T5CONCLR = 0x2; // internal pulses
  T5CONCLR = (1 << 13 | 1 << 7 | 1 << 3); // continue in idle, disable gated
  PR5 = period; // set period
  IPC5 = 0b10100; // priority 5, 0 subpriority
  IFS0bits.T5IF = 0; // clear flag
  T5CONSET = 1 << 15; // turn on timer
  IEC0bits.T5IE = 1; // enable interrupts
}


// used for setting all of the nonzero default values of 
// globals defined in MyData.h
void SetGlobalDefaults(void) {
  // default current controller gains
  Kp_curr = 0.0;
  Ki_curr = 0.0;
  // default motion control gains
  Kp = 0.0;
  Ki = 0.0;
}


// used for moving back to the beginning of a Signals struct so that we can
// start recording data again
void ResetSignal(Signals *s, int len) {
  s->index = 0;
  if (len > NMAX) {
    s->len = NMAX;
  } else {
    s->len = len;
  }
}

// Used for updating Signals structs
void UpdateSignal(Signals *s, float desired, float actual) {
  int i = s->index;

  // always update desired and actual signals, useful for printing
  // current values
  s->desired = desired;
  s->actual = actual;

  // only store if there is space
  if (i < s->len) {
    s->ref[i] = desired;
    s->act[i] = actual;
    s->index++;
  }
}
